# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Widgets for managing virtual devices

"""

import pyVmomi

import pvc.widget.common
import pvc.widget.gauge
import pvc.widget.menu
import pvc.widget.radiolist

__all__ = [
    'BaseDeviceWidget', 'AddCdromDeviceWidget',
    'AddFloppyDeviceWidget', 'AddNetworkDeviceWidget',
    'AddControllerWidget', 'AddSCSIControllerWidget',
]


class BaseDeviceWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Base device widget

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.hardware = self.obj.config.hardware
        self.display()

    def display(self):
        """
        Override this method

        """
        pass

    def choose_controller(self, controller):
        """
        Prompts the user to select a virtual controller

        Args:
            controller (vim.VirtualController): Controller type to choose from

        Returns:
            A vim.VirtualController if a virtual controller has been
            selected, None otherwise

        """
        controllers = [c for c in self.hardware.device if isinstance(c, controller)]

        if not controllers:
            self.dialog.msgbox(
                title=self.title,
                text='No suitable controllers found'
            )
            return

        items = [
            pvc.widget.menu.MenuItem(
                tag=c.deviceInfo.label,
                description='',
                on_select=lambda x: x,
                on_select_args=(c,)
            ) for c in controllers
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            return_selected=True,
            title=self.title,
            text='Select virtual controller'
        )
        item = menu.display()

        if not isinstance(item, pvc.widget.menu.MenuItem):
            return

        controller = item.selected()

        return controller

    def next_unit_number(self, controller):
        """
        Get the next unit number for a controller

        Used when adding new devices to an existing controller

        Args:
            controller (vim.VirtualController): A vim.VirtualController instance

        Returns:
            The next available unit number for the controller

        """
        used = [c.unitNumber for c in self.hardware.device if c.controllerKey == controller.key]
        unit_number = max(used) + 1 if used else 0

        return unit_number

    def next_bus_number(self, controller):
        """
        Get the next bus number for a controller

        Used when adding new virtual controllers

        Args:
            controller (vim.VirtualController): A vim.VirtualController instance

        Returns:
            The next available bus number for the controller

        """
        used = [c.busNumber for c in self.hardware.device if isinstance(c, controller)]
        bus_number = max(used) + 1 if used else 0

        return bus_number


class AddCdromDeviceWidget(BaseDeviceWidget):
    """
    Widget for adding new CD/DVD drives

    Extends:
        BaseDeviceWidget class

    Overrides:
        display() method

    """
    def display(self):
        controller = self.choose_controller(
            controller=pyVmomi.vim.VirtualIDEController
        )

        if not controller:
            return

        unit_number = self.next_unit_number(
            controller=controller
        )

        backing_info = self.select_backing()

        if not backing_info:
            return

        connect_info = pyVmomi.vim.VirtualDeviceConnectInfo(
            allowGuestControl=True,
            connected=False,
            startConnected=False
        )

        device = pyVmomi.vim.VirtualCdrom(
            backing=backing_info,
            connectable=connect_info,
            controllerKey=controller.key,
            key=-1,
            unitNumber=unit_number
        )

        device_change = pyVmomi.vim.VirtualDeviceConfigSpec(
            device=device,
            operation=pyVmomi.vim.VirtualDeviceConfigSpecOperation.add
        )

        spec = pyVmomi.vim.VirtualMachineConfigSpec(
            deviceChange=[device_change]
        )

        task = self.obj.ReconfigVM_Task(spec=spec)
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Adding CD/DVD drive ...'
        )

        gauge.display()

    def select_backing(self):
        """
        Prompts the user to select device backing

        Returns:
            A vim.VirtualDeviceDeviceBackingInfo instance on success,
            None otherwise

        """
        items = [
            pvc.widget.menu.MenuItem(
                tag='Pass Through',
                description=pyVmomi.vim.VirtualCdromRemotePassthroughBackingInfo.__name__,
                on_select=pyVmomi.vim.VirtualCdromRemotePassthroughBackingInfo,
                on_select_kwargs={'deviceName': '', 'useAutoDetect': False, 'exclusive': False}
            ),
            pvc.widget.menu.MenuItem(
                tag='ATAPI Emulation',
                description=pyVmomi.vim.VirtualCdromRemoteAtapiBackingInfo.__name__,
                on_select=pyVmomi.vim.VirtualCdromRemoteAtapiBackingInfo,
                on_select_kwargs={'deviceName': '', 'useAutoDetect': False}
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            return_selected=True,
            title=self.title,
            text='Select device backing'
        )

        item = menu.display()

        if not isinstance(item, pvc.widget.menu.MenuItem):
            self.dialog.msgbox(
                title=self.title,
                text='Invalid device backing selected'
            )
            return

        backing_info = item.selected()

        return backing_info


class AddFloppyDeviceWidget(BaseDeviceWidget):
    """
    Widget for adding new floppy drives

    Extends:
        BaseDeviceWidget class

    Overrides:
        display() method

    """
    def display(self):
        controller = self.choose_controller(
            controller=pyVmomi.vim.VirtualSIOController
        )

        if not controller:
            return

        unit_number = self.next_unit_number(
            controller=controller
        )

        backing_info = pyVmomi.vim.VirtualFloppyRemoteDeviceBackingInfo(
            deviceName='',
            useAutoDetect=False
        )

        connect_info = pyVmomi.vim.VirtualDeviceConnectInfo(
            allowGuestControl=True,
            connected=False,
            startConnected=False
        )

        device = pyVmomi.vim.VirtualFloppy(
            backing=backing_info,
            connectable=connect_info,
            controllerKey=controller.key,
            key=-1,
            unitNumber=unit_number
        )

        device_change = pyVmomi.vim.VirtualDeviceConfigSpec(
            device=device,
            operation=pyVmomi.vim.VirtualDeviceConfigSpecOperation.add
        )

        spec = pyVmomi.vim.VirtualMachineConfigSpec(
            deviceChange=[device_change]
        )

        task = self.obj.ReconfigVM_Task(spec=spec)
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Adding floppy drive ...'
        )

        gauge.display()


class AddNetworkDeviceWidget(BaseDeviceWidget):
    """
    Widget for adding network devices

    Extends:
        BaseDeviceWidget class

    Overrides:
        display() method

    """
    def display(self):
        controller = self.choose_controller(
            controller=pyVmomi.vim.VirtualPCIController
        )

        if not controller:
            return

        unit_number = self.next_unit_number(
            controller=controller
        )

        adapter = self.select_ethernet_adapter()

        if not adapter:
            return

        network = pvc.widget.common.choose_network(
            agent=self.agent,
            dialog=self.dialog,
            obj=self.obj.runtime.host
        )

        if not network:
            return

        connect_info = pyVmomi.vim.VirtualDeviceConnectInfo(
            allowGuestControl=True,
            connected=True,
            startConnected=True
        )

        backing_info = pyVmomi.vim.VirtualEthernetCardNetworkBackingInfo(
            deviceName=network.name,
            useAutoDetect=False,
            network=network
        )

        device = adapter(
            backing=backing_info,
            connectable=connect_info,
            controllerKey=controller.key,
            key=-1,
            unitNumber=unit_number
        )

        device_change = pyVmomi.vim.VirtualDeviceConfigSpec(
            device=device,
            operation=pyVmomi.vim.VirtualDeviceConfigSpecOperation.add
        )

        spec = pyVmomi.vim.VirtualMachineConfigSpec(
            deviceChange=[device_change]
        )

        task = self.obj.ReconfigVM_Task(spec=spec)
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Adding ethernet adapter ...'
        )

        gauge.display()

    def select_ethernet_adapter(self):
        """
        Prompts the user to select an ethernet adapter

        Returns:
            A vim.VirtualEthernetAdapter type on success,
            None otherwise

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        # Get the OS descriptor which contains the list of
        # supported ethernet adapters on the virtual machine
        for descriptor in self.obj.environmentBrowser.QueryConfigOption().guestOSDescriptor:
            if hasattr(descriptor, 'supportedEthernetCard'):
                break
        else:
            self.dialog.msgbox(
                title=self.title,
                text='No supported ethernet adapters found'
            )
            return

        # The on_select() callback for each MenuItem() instance that
        # we use below is used as a simple echo-like callback that
        # simply returns the selected virtual ethernet card type
        items = [
            pvc.widget.menu.MenuItem(
                tag=card.__name__.split('.')[-1].replace('Virtual', '').upper(),
                description=card.__name__,
                on_select=lambda x: x,
                on_select_args=(card,)
            ) for card in descriptor.supportedEthernetCard
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            return_selected=True,
            title=self.title,
            text='Select virtual ethernet adapter'
        )

        item = menu.display()
        if not isinstance(item, pvc.widget.menu.MenuItem):
            return

        card = item.selected()
        
        return card


class AddControllerWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Widget for adding new virtual controller

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='BusLogic Parallel',
                description='Add SCSI BusLogic Parallel Controller',
                on_select=AddSCSIControllerWidget,
                on_select_args=(self.agent, self.dialog, self.obj, pyVmomi.vim.VirtualBusLogicController)
            ),
            pvc.widget.menu.MenuItem(
                tag='LSI Logic Parallel',
                description='Add SCSI LSI Logic Parallel Controller',
                on_select=AddSCSIControllerWidget,
                on_select_args=(self.agent, self.dialog, self.obj, pyVmomi.vim.VirtualLsiLogicController)
            ),
            pvc.widget.menu.MenuItem(
                tag='LSI Logic SAS',
                description='Add SCSI LSI Logic SAS Controller',
                on_select=AddSCSIControllerWidget,
                on_select_args=(self.agent, self.dialog, self.obj, pyVmomi.vim.VirtualLsiLogicSASController)
            ),
            pvc.widget.menu.MenuItem(
                tag='VMware Paravirtual',
                description='Add SCSI VMware Paravirtual Controller',
                on_select=AddSCSIControllerWidget,
                on_select_args=(self.agent, self.dialog, self.obj, pyVmomi.vim.ParaVirtualSCSIController)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select a virtual controller to add'
        )

        menu.display()


class AddSCSIControllerWidget(BaseDeviceWidget):
    """
    Widget for adding virtual SCSI controller

    Extends:
        BaseDeviceWidget class

    Overrides:
        display() method

    """
    def __init__(self, agent, dialog, obj, scsi_controller):
        """
        SCSI device widget

        Args:
            agent                          (VConnector): A VConnector instance
            dialog                      (dialog.Dialog): A Dialog instance
            obj                    (vim.VirtualMachine): A VirtualMachine managed entity
            scsi_controller (vim.VirtualSCSIController): A SCSI controller type

        """
        self.scsi_controller = scsi_controller
        super(AddSCSIControllerWidget, self).__init__(agent, dialog, obj)

    def display(self):
        controller = self.choose_controller(
            controller=pyVmomi.vim.VirtualPCIController
        )

        if not controller:
            return

        unit_number = self.next_unit_number(controller=controller)
        bus_number = self.next_bus_number(controller=self.scsi_controller)
        device = self.scsi_controller(
            controllerKey=controller.key,
            key=-1,
            unitNumber=unit_number,
            busNumber=bus_number,
            hotAddRemove=True,
            sharedBus=pyVmomi.vim.VirtualSCSISharing.noSharing
        )

        device_change = pyVmomi.vim.VirtualDeviceConfigSpec(
            device=device,
            operation=pyVmomi.vim.VirtualDeviceConfigSpecOperation.add
        )

        spec = pyVmomi.vim.VirtualMachineConfigSpec(
            deviceChange=[device_change]
        )

        task = self.obj.ReconfigVM_Task(spec=spec)
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Adding SCSI controller ...'
        )

        gauge.display()



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Virtual Machine Widgets

"""

import os
import platform
import time
import tarfile

import pyVmomi
import humanize
import requests

import pvc.widget.alarm
import pvc.widget.common
import pvc.widget.device
import pvc.widget.debug
import pvc.widget.event
import pvc.widget.menu
import pvc.widget.form
import pvc.widget.gauge
import pvc.widget.vnc
import pvc.widget.network
import pvc.widget.performance
import pvc.widget.radiolist

from subprocess import Popen, PIPE

__all__ = [
    'VirtualMachineWidget',
    'VirtualMachineActionWidget',
    'VirtualMachineConsoleWidget',
    'VirtualMachinePowerWidget',
    'VirtualMachineExportWidget',
    'CreateVirtualMachineWidget',
    'VirtualMachineHardwareWidget',
    'VirtualMachineAddHardwareWidget',
    'MigrateVirtualMachineWidget',
    'VirtualMachineChangeHostWidget',
    'VirtualMachineChangeDatastoreWidget',
    'VirtualMachineCloneWidget',
]


class VirtualMachineWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Virtual Machine Widget

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='General',
                description='General information',
                on_select=self.general_info
            ),
            pvc.widget.menu.MenuItem(
                tag='Resources',
                description='Resources usage information ',
                on_select=self.resources_info
            ),
            pvc.widget.menu.MenuItem(
                tag='Actions',
                description='Available Actions',
                on_select=VirtualMachineActionWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Migrate',
                description='Migrate Virtual Machine',
                on_select=MigrateVirtualMachineWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Clone',
                description='Clone Virtual Machine',
                on_select=VirtualMachineCloneWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Power',
                description='Virtual Machine Power Options',
                on_select=VirtualMachinePowerWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Hardware',
                description='Manage virtual hardware',
                on_select=VirtualMachineHardwareWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Datastore',
                description='Datastores used by the VM',
                on_select=pvc.widget.common.datastore_menu,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Network',
                description='Virtual Machine Networking',
                on_select=pvc.widget.common.network_menu,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Performance',
                description='Performance Metrics',
                on_select=pvc.widget.performance.PerformanceProviderWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Template',
                description='Template Actions',
                on_select=VirtualMachineTemplateWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Events',
                description='View Events',
                on_select=pvc.widget.event.EventWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Alarms',
                description='View triggered alarms',
                on_select=pvc.widget.common.alarm_menu,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Console',
                description='Launch Console',
                on_select=VirtualMachineConsoleWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Debug',
                description='Start a Python REPL console',
                on_select=pvc.widget.debug.DebugWidget,
                on_select_args=(locals(), globals())
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an item from the menu'
        )

        menu.display()

    def general_info(self):
        """
        Virtual Machine general information

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        elements = [
            pvc.widget.form.FormElement(
                label='Guest OS',
                item=self.obj.config.guestFullName if self.obj.config.guestFullName else 'Unknown'
            ),
            pvc.widget.form.FormElement(
                label='VM Version',
                item=self.obj.config.version
            ),
            pvc.widget.form.FormElement(
                label='CPU',
                item='{} vCPU(s)'.format(self.obj.config.hardware.numCPU)
            ),
            pvc.widget.form.FormElement(
                label='Memory',
                item='{} MB'.format(self.obj.config.hardware.memoryMB)
            ),
            pvc.widget.form.FormElement(
                label='Memory Overhead',
                item='{} MB'.format(self.obj.summary.quickStats.consumedOverheadMemory)
            ),
            pvc.widget.form.FormElement(
                label='VMware Tools Status',
                item=self.obj.guest.toolsRunningStatus
            ),
            pvc.widget.form.FormElement(
                label='VMware Tools Version',
                item=self.obj.guest.toolsVersionStatus
            ),
            pvc.widget.form.FormElement(
                label='IP Address',
                item=self.obj.guest.ipAddress if self.obj.guest.ipAddress else 'Unknown'
            ),
            pvc.widget.form.FormElement(
                label='DNS Name',
                item=self.obj.guest.hostName if self.obj.guest.hostName else 'Unknown'
            ),
            pvc.widget.form.FormElement(
                label='State',
                item=self.obj.runtime.powerState
            ),
            pvc.widget.form.FormElement(
                label='Host',
                item=self.obj.runtime.host.name
            ),
            pvc.widget.form.FormElement(
                label='Template',
                item=str(self.obj.config.template)
            ),
            pvc.widget.form.FormElement(
                label='Folder',
                item=self.obj.parent.name
            ),
            pvc.widget.form.FormElement(
                label='VMX Path',
                item=self.obj.config.files.vmPathName
            ),
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title=self.title,
            text='Virtual Machine general information'
        )

        form.display()

    def resources_info(self):
        """
        Resources usage information

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        provisioned_storage = self.obj.summary.storage.committed + \
            self.obj.summary.storage.uncommitted

        elements = [
            pvc.widget.form.FormElement(
                label='Consumed Host CPU',
                item='{} MHz'.format(self.obj.summary.quickStats.overallCpuUsage)
            ),
            pvc.widget.form.FormElement(
                label='Consumed Host Memory',
                item='{} MB'.format(self.obj.summary.quickStats.hostMemoryUsage)
            ),
            pvc.widget.form.FormElement(
                label='Active Guest Memory',
                item='{} MB'.format(self.obj.summary.quickStats.guestMemoryUsage)
            ),
            pvc.widget.form.FormElement(
                label='Provisioned Storage',
                item=humanize.naturalsize(provisioned_storage, binary=True)
            ),
            pvc.widget.form.FormElement(
                label='Non-shared Storage',
                item=humanize.naturalsize(self.obj.summary.storage.unshared, binary=True)
            ),
            pvc.widget.form.FormElement(
                label='Used Storage',
                item=humanize.naturalsize(self.obj.summary.storage.committed, binary=True)
            ),
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title=self.title,
            text='Virtual Machine resource usage'
        )

        return form.display()


class VirtualMachinePowerWidget(object):
    """
    Virtual Machine Power Menu Widget

    """
    def __init__(self, agent, dialog, obj):
        """
        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Power On',
                description='Power On Virtual Machine',
                on_select=self.power_on
            ),
            pvc.widget.menu.MenuItem(
                tag='Power Off',
                description='Power Off Virtual Machine Off ',
                on_select=self.power_off
            ),
            pvc.widget.menu.MenuItem(
                tag='Suspend',
                description='Suspend Virtual Machine',
                on_select=self.suspend,
            ),
            pvc.widget.menu.MenuItem(
                tag='Reset',
                description='Reset Virtual Machine',
                on_select=self.reset
            ),
            pvc.widget.menu.MenuItem(
                tag='Shutdown',
                description='Shutdown Guest System',
                on_select=self.shutdown
            ),
            pvc.widget.menu.MenuItem(
                tag='Reboot',
                description='Reboot Guest System',
                on_select=self.reboot
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()

    def power_on(self):
        """
        Power on the virtual machine

        """
        if self.obj.runtime.powerState == pyVmomi.vim.VirtualMachinePowerState.poweredOn:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine is already powered on.'
            )
            return

        task = self.obj.PowerOn()
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Powering On Virtual Machine'
        )

        gauge.display()

    def power_off(self):
        """
        Power off the virtual machine

        """
        if self.obj.runtime.powerState == pyVmomi.vim.VirtualMachinePowerState.poweredOff:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine is already powered off.'
            )
            return

        task = self.obj.PowerOff()
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Powering Off Virtual Machine'
        )

        gauge.display()

    def suspend(self):
        """
        Suspend the virtual machine

        """
        if self.obj.runtime.powerState != pyVmomi.vim.VirtualMachinePowerState.poweredOn:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine is not powered on, cannot suspend.'
            )
            return

        task = self.obj.Suspend()
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Suspending Virtual Machine'
        )

        gauge.display()

    def reset(self):
        """
        Reset the virtual machine

        """
        if self.obj.runtime.powerState != pyVmomi.vim.VirtualMachinePowerState.poweredOn:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine is not powered on, cannot reset.'
            )
            return

        task = self.obj.Reset()
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Resetting Virtual Machine'
        )

        gauge.display()

    def shutdown(self):
        """
        Shutdown the virtual machine

        For a proper guest shutdown we need VMware Tools running

        """
        if self.obj.runtime.powerState != pyVmomi.vim.VirtualMachinePowerState.poweredOn:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine is not powered on, cannot shutdown.'
            )
            return

        if self.obj.guest.toolsRunningStatus != pyVmomi.vim.VirtualMachineToolsRunningStatus.guestToolsRunning:
            self.dialog.msgbox(
                title=self.title,
                text='VMware Tools is not running, cannot shutdown system'
            )
            return

        self.dialog.infobox(
            title=self.title,
            text='Shutting down guest system ...'
        )
        self.obj.ShutdownGuest()

    def reboot(self):
        """
        Reboot the virtual machine

        For a proper guest reboot we need VMware Tools running

        """
        if self.obj.runtime.powerState != pyVmomi.vim.VirtualMachinePowerState.poweredOn:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine is not powered on, cannot reboot.'
            )
            return

        if self.obj.guest.toolsRunningStatus != pyVmomi.vim.VirtualMachineToolsRunningStatus.guestToolsRunning:
            self.dialog.msgbox(
                title=self.title,
                text='VMware Tools is not running, cannot reboot system'
            )
            return

        self.dialog.infobox(
            title=self.title,
            text='Rebooting guest system ...'
        )
        self.obj.RebootGuest()


class VirtualMachineExportWidget(object):
    def __init__(self, agent, dialog, obj, create_ova):
        """
        Virtual Machine Export Widget

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity
            create_ova           (bool): If True then export VM into a single OVA file
                                         Otherwise create a folder of files (OVF)

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.create_ova = create_ova
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        if self.obj.runtime.powerState != pyVmomi.vim.VirtualMachinePowerState.poweredOff:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine must be powered off in order to be exported'
            )
            return

        code, path = self.dialog.dselect(
            title='Directory to save OVF template',
            filepath=''
        )

        path = os.path.join(path, self.obj.name)

        if code in (self.dialog.ESC, self.dialog.CANCEL):
            self.dialog.msgbox(
                title=self.title,
                text='No destination directory specified'
            )
            return

        if not os.path.exists(path):
            os.makedirs(path)

        self.export_ovf_template(path=path)

    def export_ovf_template(self, path):
        """
        Exports a Virtual Machine into OVF/OVA template

        Args:
            path (str): Directory to save the OVF/OVA template

        """
        # TODO: Perform a dry-run and see if creating the
        #       OVF descriptor succeeds and then proceed with
        #       downloading the actual VMDK files

        self.dialog.infobox(
            title=self.title,
            text='Initializing OVF export ...'
        )

        lease = self.obj.ExportVm()

        while True:
            if lease.state == pyVmomi.vim.HttpNfcLeaseState.initializing:
                lease.HttpNfcLeaseProgress(percent=0)
            elif lease.state == pyVmomi.vim.HttpNfcLeaseState.error:
                self.dialog.msgbox(
                    title=self.obj.name,
                    text=lease.error.msg
                )
                lease.HttpNfcLeaseAbort()
                return
            elif lease.state == pyVmomi.vim.HttpNfcLeaseState.ready:
                break
            time.sleep(0.5)

        percent = 0
        total_transfered_bytes = 0
        manifest = []
        ovf_files = []
        exported_disks = {}

        self.dialog.gauge_start(
            title='Exporting OVF template - {}'.format(self.obj.name)
        )

        for url in lease.info.deviceUrl:
            if not url.disk:  # skip non-vmdk disks
                continue

            self.dialog.gauge_update(
                percent=percent,
                text='Exporting {} ...\n'.format(url.targetId),
                update_text=True
            )

            if manifest:
                total_transfered_bytes = sum([m.capacity for m in manifest])

            disk_transfered_bytes = 0
            disk_file = os.path.join(path, '{}-{}'.format(self.obj.name, url.targetId))
            r = requests.get(url.url, verify=False, stream=True)

            with open(disk_file, 'wb') as f:
                for chunk in r.iter_content(chunk_size=512*1024):
                    if chunk:
                        disk_transfered_bytes += len(chunk)
                        percent = round(
                            (total_transfered_bytes + disk_transfered_bytes) / 1024 /
                            lease.info.totalDiskCapacityInKB * 100
                        )
                        self.dialog.gauge_update(percent=percent)
                        lease.HttpNfcLeaseProgress(percent=percent)
                        f.write(chunk)

            m = [me for me in lease.HttpNfcLeaseGetManifest() if me.key == url.key].pop()
            manifest.append(m)

            of = pyVmomi.vim.OvfManager.OvfFile(
                capacity=m.capacity,
                deviceId=m.key,
                path=os.path.basename(disk_file),
                populatedSize=m.populatedSize,
                size=disk_transfered_bytes,
            )
            ovf_files.append(of)

            exported_disks[url.key] = url.targetId
            total_transfered_bytes = sum([me.capacity for me in manifest])
            percent = round(
                total_transfered_bytes / 1024 / lease.info.totalDiskCapacityInKB * 100
            )
            self.dialog.gauge_update(percent)
            lease.HttpNfcLeaseProgress(percent=percent)

        # Create OVF manifest and descriptor files
        self.dialog.gauge_stop()
        self.create_manifest_file(
            path=path,
            manifest=manifest,
            disks=exported_disks
        )

        self.create_ovf_descriptor(
            path=path,
            ovf_files=ovf_files
        )

        lease.HttpNfcLeaseComplete()

        if self.create_ova:
            self.create_ova_file(
                path=path,
                disks=exported_disks.values()
            )

        self.dialog.msgbox(
            title=self.title,
            text='Export successful. Files saved in:\n\n{}\n'.format(path)
        )

    def create_manifest_file(self, path, manifest, disks):
        """
        Creates the OVF manifest file

        Args:
            path      (str): Path to the exported disks
            manifest (list): A list of vim.HttpNfcLease.ManifestEntry instances
            disks    (dict): A mapping of the disk keys and target ids

        """
        self.dialog.infobox(
            title=self.title,
            text='Creating OVF manifest ...'
        )

        with open(os.path.join(path, '{}.mf'.format(self.obj.name)), 'w') as f:
            for entry in manifest:
                f.write('SHA1({}-{})= {}\n'.format(
                    self.obj.name,
                    disks[entry.key],
                    entry.sha1)
                )

    def create_ovf_descriptor(self, path, ovf_files):
        """
        Creates the OVF descriptor file

        Args:
            path       (str): Path to the exported disks
            ovf_files (list): A list of vim.OvfManager.OvfFile instances

        """
        self.dialog.infobox(
            title=self.title,
            text='Creating OVF descriptor ...'
        )

        cdp = pyVmomi.vim.OvfManager.CreateDescriptorParams(
            ovfFiles=ovf_files
        )

        dr = self.agent.si.content.ovfManager.CreateDescriptor(
            obj=self.obj,
            cdp=cdp
        )

        if dr.warning:
            self.dialog.msgbox(
                title='Warning - {}'.format(self.title),
                text=str(dr.warning)
            )

        if dr.error:
            self.dialog.msgbox(
                title='Error - {}'.format(self.title),
                text=str(dr.error)
            )

        with open(os.path.join(path, '{}.ovf'.format(self.obj.name)), 'w') as f:
            f.write(dr.ovfDescriptor)

    def create_ova_file(self, path, disks):
        """
        Creates a single OVA file of the exported VM

        Args:
            path   (str): Path to the exported disks
            disks (list): A list of the downloaded disks

        """
        self.dialog.infobox(
            title=self.title,
            text='Creating OVA file ...'
        )

        old_cwd = os.getcwd()
        os.chdir(path)

        ova = tarfile.open('{}.ova'.format(self.obj.name), 'w')

        # Add descriptor and manifest files first
        descriptor = '{}.ovf'.format(self.obj.name)
        manifest = '{}.mf'.format(self.obj.name)
        ova.add(descriptor)
        ova.add(manifest)

        # Now add the VMDK disks
        for disk in disks:
            ova.add('{}-{}'.format(self.obj.name, disk))

        ova.close()

        # Cleanup disks, manifest and descriptor files
        os.unlink(manifest)
        os.unlink(descriptor)
        for disk in disks:
            os.unlink('{}-{}'.format(self.obj.name, disk))

        os.chdir(old_cwd)


class VirtualMachineConsoleWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Virtual Machine Console Widget

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='VNC',
                description='Launch VNC Console',
                on_select=pvc.widget.vnc.VncWidget,
                on_select_args=(self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='VMRC',
                description='Launch VMRC Console',
                on_select=self.console,
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select a console to be launched'
        )

        menu.display()

    def console(self):
        """
        Launch a VMRC console to the Virtual Machine

        On GNU/Linux systems we use VMware Player for
        launching a console to the Virtual Machine.

        On Mac OS X we use VMRC for launching the console.

        VMware Player console is launched using this syntax:

            $ vmplayer -h <hostname> -p <ticket> -M <managed-object-id>

        VMRC console is launched using this syntax:

            $ vmrc vmrc://clone:<ticket>@<hostname>/?moid=<managed-object-id>

        Where <ticket> is an acquired ticket as returned by a
        previous call to AcquireCloneTicket().

        TODO: Drop support for VMware Player on GNU/Linux as soon as
              support for VMRC is provided for GNU/Linux systems.
              See http://kb.vmware.com/kb/2091284 for more details

        """
        self.dialog.infobox(
            title=self.title,
            text='Launching console ...'
        )

        ticket = self.agent.si.content.sessionManager.AcquireCloneTicket()

        try:
            if platform.system() == 'Darwin':
                args=['open', 'vmrc://clone:{}@{}/?moid={}'.format(
                                                                ticket,
                                                                self.agent.host,
                                                                self.obj._moId)]
            else:
                args=['vmplayer', '-h', self.agent.host, '-p', ticket, '-M', self.obj._moId]
            Popen(
                args=args,
                stdout=PIPE,
                stderr=PIPE
            )
        except OSError as e:
            self.dialog.msgbox(
                title=self.title,
                text='Cannot launch console: \n{}\n'.format(e)
            )
            return

        # Give it some time to start up the console
        time.sleep(3)


class VirtualMachineTemplateWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Virtual Machine Template Widget

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Mark Template',
                description='Mark Virtual Machine as template',
                on_select=self.mark_as_template
            ),
            pvc.widget.menu.MenuItem(
                tag='Mark VM',
                description='Mark template as Virtual Machine',
                on_select=self.mark_as_virtual_machine
            ),
            pvc.widget.menu.MenuItem(
                tag='Export OVA',
                description='Export as single file (OVA)',
                on_select=VirtualMachineExportWidget,
                on_select_args=(self.agent, self.dialog, self.obj, True)
            ),
            pvc.widget.menu.MenuItem(
                tag='Export OVF',
                description='Export as directory of files (OVF)',
                on_select=VirtualMachineExportWidget,
                on_select_args=(self.agent, self.dialog, self.obj, False)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            dialog=self.dialog,
            items=items,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()

    def mark_as_template(self):
        """
        Mark Virtual Machine as a template

        """
        if self.obj.runtime.powerState != pyVmomi.vim.VirtualMachinePowerState.poweredOff:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine must be powered off first'
            )
            return

        self.dialog.infobox(
            title=self.title,
            text='Marking {} as template ...'.format(self.obj.name)
        )

        try:
            self.obj.MarkAsTemplate()
        except pyVmomi.vim.MethodFault as e:
            self.dialog.msgbox(
                title=self.title,
                text=e.msg
            )

    def mark_as_virtual_machine(self):
        """
        Marks a template as a Virtual Machine

        """
        self.dialog.infobox(
            title=self.title,
            text='Marking {} as virtual machine ...'.format(self.obj.name)
        )

        try:
            self.obj.MarkAsVirtualMachine()
        except pyVmomi.vim.MethodFault as e:
            self.dialog.msgbox(
                title=self.title,
                text=e.msg
            )


class VirtualMachineActionWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Virtual Machine Action Widget

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Rename',
                description='Rename Virtual Machine',
                on_select=pvc.widget.common.rename,
                on_select_args=(self.obj, self.dialog)
            ),
            pvc.widget.menu.MenuItem(
                tag='Unregister',
                description='Remove from inventory',
                on_select=self.unregister
            ),
            pvc.widget.menu.MenuItem(
                tag='Delete',
                description='Delete from disk',
                on_select=pvc.widget.common.remove,
                on_select_args=(self.obj, self.dialog)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            dialog=self.dialog,
            items=items,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()

    def unregister(self):
        """
        Unregister the VM from inventory

        """
        code = self.dialog.yesno(
            title='Confirm Remove',
            text='Remove {} from inventory?'.format(self.obj.name)
        )

        if code in (self.dialog.ESC, self.dialog.CANCEL):
            return

        self.dialog.infobox(
            title=self.title,
            text='Removing {} from inventory ...'.format(self.obj.name)
        )

        self.obj.UnregisterVM()


class CreateVirtualMachineWidget(object):
    def __init__(self, agent, dialog, datacenter=None, cluster=None, host=None):
        """
        Widget for creating a new Virtual Machine

        Args:
            agent                      (VConnector): A VConnector instance
            dialog                  (dialog.Dialog): A Dialog instance
            datacenter             (vim.Datacenter): A vim.Datacenter instance
            cluster    (vim.ClusterComputeResource): A vim.CluterComputeResource instance
            host                   (vim.HostSystem): A vim.HostSystem instance

        """
        self.agent = agent
        self.dialog = dialog
        self.datacenter = datacenter
        self.cluster = cluster
        self.host = host
        self.display()

    def display(self):
        if not self.datacenter:
            self.datacenter = self.select_datacenter()
            if not self.datacenter:
                return

        if not self.cluster:
            self.cluster = self.select_cluster(folder=self.datacenter)
            if not self.cluster:
                return

        if not self.select_host(cluster=self.cluster):
            return

        if self.host:
            datastore = self.select_datastore(obj=self.host)
        else:
            datastore = self.select_datastore(obj=self.cluster)
        if not datastore:
            return

        vmx_version = self.select_vmx_version(obj=self.cluster)
        if not vmx_version:
            return

        # TODO: Add guest operation system
        # TODO: Add virtual disks
        # TODO: Add networking

        specs = self.get_vm_specs()
        if not specs:
            return

        folder = self.datacenter.vmFolder
        pool = self.cluster.resourcePool

        vmx_file = pyVmomi.vim.VirtualMachineFileInfo(
            vmPathName='[{}] {}'.format(datastore.name, specs['Name'])
        )

        config_spec = pyVmomi.vim.VirtualMachineConfigSpec(
            name=specs['Name'],
            numCPUs=int(specs['vCPU(s)']),
            memoryMB=int(specs['Memory Size (MB)']),
            guestId=specs['Guest ID'],
            files=vmx_file,
            version=vmx_version
        )

        task = folder.CreateVM_Task(
            config=config_spec,
            pool=pool,
            host=self.host
        )
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title='Creating New Virtual Machine',
            text='Creating virtual machine {}'.format(specs['Name'])
        )

        gauge.display()

    def select_datacenter(self):
        """
        Select datacenter for Virtual Machine placement

        Returns:
            A vim.Datacenter managed entity upon successfuly selecting
            an existing vim.Datacenter instance, None otherwise

        """
        datacenter = pvc.widget.common.choose_datacenter(
            agent=self.agent,
            dialog=self.dialog
        )

        if not datacenter:
            return

        if not datacenter.hostFolder.childEntity:
            self.dialog.msgbox(
                title='Create New Virtual Machine',
                text='No compute resources found in datacenter {}'.format(datacenter.name)
            )
            return

        return datacenter

    def select_cluster(self, folder):
        """
        Select a cluster for Virtual Machine placement

        Args:
            folder (vim.Folder): A vim.Folder instance

        Returns:
            A vim.ClusterComputeResource managed entity upon successufuly
            selecting an existing cluster, None otherwise

        """
        cluster = pvc.widget.common.choose_cluster(
            agent=self.agent,
            dialog=self.dialog,
            folder=folder
        )

        if not cluster:
            return

        if not cluster.host:
            self.dialog.msgbox(
                title='Create New Virtual Machine',
                text='No valid hosts found in cluster {}'.format(cluster.name)
            )
            return

        return cluster

    def select_host(self, cluster):
        """
        Select host for Virtual Machine placement

        If a target host has been provided use that host for the
        Virtual Machine placement.

        If the cluster has DRS enabled in Fully Automated mode then
        let vSphere decide where to place the Virtual Machine instead.

        Args:
            cluster (vim.ClusterComputeResource): A cluster managed entity

        Returns:
            True on successfully selecting a host or if the cluster is
            configured with DRS in fully automated mode allowing for
            dynamic Virtual Machine placement, otherwise returns False.

        """
        drs_config = cluster.configuration.drsConfig
        drs_enabled = drs_config.enabled
        drs_fully_automated = drs_config.defaultVmBehavior == pyVmomi.vim.cluster.DrsConfigInfo.DrsBehavior.fullyAutomated
        dynamic_placement = drs_enabled and drs_fully_automated

        if not self.host and not dynamic_placement:
            self.host = pvc.widget.common.choose_host(
                agent=self.agent,
                dialog=self.dialog,
                folder=cluster
            )

            if not self.host:
                self.dialog.msgbox(
                    title='Create New Virtual Machine',
                    text='No valid host selected'
                )
                return False

        return True

    def select_datastore(self, obj):
        """
        Select datastore for Virtual Machine placement

        Args:
            obj (vim.ManagedEntity): A Managed Entity containing datastores

        Returns:
            A vim.Datastore managed entity upon successfully
            selecting a datastore, None otherwise

        """
        if not obj.datastore:
            self.dialog.msgbox(
                title='Create New Virtual Machine',
                text='No datastores found on {}'.format(obj.name)
            )
            return

        datastore = pvc.widget.common.choose_datastore(
            agent=self.agent,
            dialog=self.dialog,
            obj=obj
        )

        return datastore

    def select_vmx_version(self, obj):
        """
        Select a VMX version for the Virtual Machine

        Args:
            obj (vim.ComputeResource): Entity to query for supported versions

        """
        self.dialog.infobox(
            text='Retrieving information ...'
        )

        versions = obj.environmentBrowser.QueryConfigOptionDescriptor()
        items = [
            pvc.widget.radiolist.RadioListItem(
                tag=v.key,
                description=v.description
            ) for v in versions if v.createSupported
        ]

        radiolist = pvc.widget.radiolist.RadioList(
            items=items,
            dialog=self.dialog,
            title='Create New Virtual Machine',
            text='Select Virtual Machine hardware version'
        )

        code, tag = radiolist.display()

        if code in (self.dialog.CANCEL, self.dialog.ESC) or not tag:
            return

        return tag

    def get_vm_specs(self):
        """
        Get Virtual Machine specs

        """
        elements = [
            pvc.widget.form.FormElement(label='Name'),
            pvc.widget.form.FormElement(label='vCPU(s)'),
            pvc.widget.form.FormElement(label='Memory Size (MB)'),
            pvc.widget.form.FormElement(label='Guest ID', item='otherGuest64'),
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title='Create New Virtual Machine',
            text='Virtual Machine details'
        )

        code, fields = form.display()

        if code in (self.dialog.CANCEL, self.dialog.ESC):
            return

        if not all(fields.values()):
            self.dialog.msgbox(
                title='Create New Virtual Machine',
                text='Invalid input provided'
            )
            return

        return fields


class VirtualMachineHardwareWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Widget for managing hardware on a Virtual Machine

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Add',
                description='Add virtual hardware',
                on_select=VirtualMachineAddHardwareWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Remove',
                description='Remove virtual hardware'
            ),
            pvc.widget.menu.MenuItem(
                tag='Change',
                description='Change virtual hardware'
            ),
            pvc.widget.menu.MenuItem(
                tag='View',
                description='View virtual hardware',
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()


class VirtualMachineAddHardwareWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Widget for adding new virtual hardware

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Floppy Drive',
                description='Add floppy drive',
                on_select=pvc.widget.device.AddFloppyDeviceWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='CD/DVD Drive',
                description='Add CD/DVD drive',
                on_select=pvc.widget.device.AddCdromDeviceWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Hard Disk',
                description='Add hard disk'
            ),
            pvc.widget.menu.MenuItem(
                tag='Ethernet Adapter',
                description='Add ethernet adapter',
                on_select=pvc.widget.device.AddNetworkDeviceWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Controller',
                description='Add virtual controller',
                on_select=pvc.widget.device.AddControllerWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()


class MigrateVirtualMachineWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Widget for migrating a virtual machine

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Change Host',
                description='Migrate VM to another host',
                on_select=VirtualMachineChangeHostWidgetWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Change Datastore',
                description='Relocate VM to another datastore',
                on_select=VirtualMachineChangeDatastoreWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()


class VirtualMachineChangeHostWidgetWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Widget for migrating a virtual machine to another host

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        if self.obj.config.template:
            self.dialog.msgbox(
                title=self.title,
                text='Virtual Machine is a template, cannot migrate'
            )
            return

        folder = self.obj.runtime.host.parent.parent
        cluster = pvc.widget.common.choose_cluster(
            agent=self.agent,
            dialog=self.dialog,
            folder=folder
        )

        if not cluster:
            return

        if not cluster.host:
            self.dialog.msgbox(
                title=self.title,
                text='No compute resources found in {}'.format(cluster.name)
            )
            return

        old_host = self.obj.runtime.host
        new_host = pvc.widget.common.choose_host(
            agent=self.agent,
            dialog=self.dialog,
            folder=cluster
        )

        if not new_host:
            self.dialog.msgbox(
                title=self.title,
                text='No valid host selected'
            )
            return

        if not self.migration_would_succeed(host=new_host, pool=cluster.resourcePool):
            return

        task = self.obj.MigrateVM_Task(
            host=new_host,
            priority=pyVmomi.vim.VirtualMachineMovePriority.defaultPriority
        )

        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Migrating Virtual Machine from {} to {} ...'.format(
                old_host.name,
                new_host.name
            )
        )

        gauge.display()

    def migration_would_succeed(self, host, pool):
        """
        Runs compatibility tests in order to determine whether a
        migration task would succeed or not.

        Args:
            host   (vim.HostSystem): A target HostSystem entity to migrate the VM on
            pool (vim.ResourcePool): A target ResourcePool entity for the VM

        Returns:
            True if there are no errors reported by the compatibily tests,
            False otherwise.

        """
        task = self.agent.si.content.vmProvisioningChecker.CheckMigrate_Task(
            vm=self.obj,
            host=host,
            pool=pool
        )

        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Running compatibility tests ...'
        )

        gauge.display()

        warnings = [r for r in task.info.result if r.warning]
        errors = [r for r in task.info.result if r.error]

        warning_messages = []
        for r in warnings:
            for w in r.warning:
                warning_messages.append(w.msg)
                warning_messages.extend(['    {}'.format(f.message) for f in w.faultMessage])

        error_messages = []
        for r in errors:
            for e in r.error:
                error_messages.append(e.msg)
                error_messages.extend(['    {}'.format(f.message) for f in e.faultMessage])

        if warning_messages:
            warn_msg = (
                'The following warnings were reported by the '
                'compatibility tests while testing the '
                'feasibility of the proposed migration task.\n\n'
                '{}'
            )

            self.dialog.msgbox(
                title='Migration Task Warning',
                text=warn_msg.format('\n'.join(warning_messages))
            )

        if error_messages:
            error_msg = (
                'The following errors were reported by the '
                'compatibility tests while testing the '
                'feasibility of the proposed migration task.\n\n'
                'As a result of these errors the migration task '
                'would not succeed. You are advised to check and '
                'fix these errors before attempting to migrate the '
                'Virtual Machine.\n\n'
                '{}'
            )

            self.dialog.msgbox(
                title='Migration Task Error',
                text=error_msg.format('\n'.join(error_messages))
            )
            return False  # Migration would not succeed

        return True


class VirtualMachineChangeDatastoreWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Widget for relocating a virtual machine to another datastore

        Args:
            agent          (VConnector): A VConnector instance
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        new_datastore = pvc.widget.common.choose_datastore(
            agent=self.agent,
            dialog=self.dialog,
            obj=self.obj.runtime.host
        )

        if not new_datastore:
            self.dialog.msgbox(
                title=self.title,
                text='No target datastore selected'
            )
            return

        if not self.relocation_would_succeed(datastore=new_datastore):
            return

        spec = pyVmomi.vim.VirtualMachineRelocateSpec(
            datastore=new_datastore
        )

        task = self.obj.RelocateVM_Task(
            spec=spec,
            priority=pyVmomi.vim.VirtualMachineMovePriority.defaultPriority
        )

        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Relocating Virtual Machine to {} ...'.format(new_datastore.name)
        )

        gauge.display()

    def relocation_would_succeed(self, datastore):
        """
        Runs compatibility tests in order to determine whether a
        relocation task would succeed or not.

        Args:
            datastore (vim.Datastore): Target Datastore entity to relocate the VM on

        Returns:
            True if there are no errors reported by the compatibily tests,
            False otherwise.

        """
        spec = pyVmomi.vim.VirtualMachineRelocateSpec(
            datastore=datastore
        )

        task = self.agent.si.content.vmProvisioningChecker.CheckRelocate_Task(
            vm=self.obj,
            spec=spec
        )

        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Running compatibility tests ...'
        )

        gauge.display()

        warnings = [r for r in task.info.result if r.warning]
        errors = [r for r in task.info.result if r.error]

        warning_messages = []
        for r in warnings:
            for w in r.warning:
                warning_messages.append(w.msg)
                warning_messages.extend(['    {}'.format(f.message) for f in w.faultMessage])

        error_messages = []
        for r in errors:
            for e in r.error:
                error_messages.append(e.msg)
                error_messages.extend(['    {}'.format(f.message) for f in e.faultMessage])

        if warning_messages:
            warn_msg = (
                'The following warnings were reported by the '
                'compatibility tests while testing the '
                'feasibility of the proposed relocation task.\n\n'
                '{}'
            )

            self.dialog.msgbox(
                title='Relocation Task Warning',
                text=warn_msg.format('\n'.join(warning_messages))
            )

        if error_messages:
            error_msg = (
                'The following errors were reported by the '
                'compatibility tests while testing the '
                'feasibility of the proposed relocation task.\n\n'
                'As a result of these errors the relocation task '
                'would not succeed. You are advised to check and '
                'fix these errors before attempting to relocate the '
                'Virtual Machine.\n\n'
                '{}'
            )

            self.dialog.msgbox(
                title='Relocation Task Error',
                text=error_msg.format('\n'.join(error_messages))
            )
            return False  # Relocation would not succeed

        return True


class VirtualMachineCloneWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Widget for cloning a Virtual Machine

        Args:
            agent           (VConnector): A VConnector instance
            dialog       (dialog.Dialog): A Dialog instance
            obj     (vim.VirtualMachine): A VirtualMachine managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        code, name = self.dialog.inputbox(
            title='Clone Virtual Machine',
            text='Name of the target Virtual Machine clone'
        )

        if code in (self.dialog.CANCEL, self.dialog.ESC):
            return

        datacenter = self.select_datacenter()
        if not datacenter:
            return

        cluster = self.select_cluster(folder=datacenter)
        if not cluster:
            return

        drs_config = cluster.configuration.drsConfig
        drs_enabled = drs_config.enabled
        drs_fully_automated = drs_config.defaultVmBehavior == pyVmomi.vim.cluster.DrsConfigInfo.DrsBehavior.fullyAutomated
        dynamic_placement = drs_enabled and drs_fully_automated

        host = None
        if not dynamic_placement:
            host = pvc.widget.common.choose_host(
                agent=self.agent,
                dialog=self.dialog,
                folder=cluster
            )

            if not host:
                self.dialog.msgbox(
                    title='Clone Virtual Machine',
                    text='No valid host selected'
                )
                return

        if not dynamic_placement:
            datastore = self.select_datastore(obj=host)
        else:
            datastore = self.select_datastore(obj=cluster)

        if not datastore:
            return

        folder = datacenter.vmFolder
        pool = cluster.resourcePool
        relocate_spec = pyVmomi.vim.VirtualMachineRelocateSpec(
            datastore=datastore,
            host=host,
            pool=pool
        )

        clone_spec = pyVmomi.vim.VirtualMachineCloneSpec(
            location=relocate_spec,
            powerOn=False,
            template=False
        )

        task = self.obj.CloneVM_Task(
            folder=folder,
            name=name,
            spec=clone_spec
        )

        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title='Clone Virtual Machine {}'.format(self.title),
            text='Cloning virtual machine {} to {} ...'.format(self.obj.name, name)
        )

        gauge.display()

    def select_datacenter(self):
        """
        Select datacenter for Virtual Machine clone placement

        Returns:
            A vim.Datacenter managed entity upon successfuly selecting
            an existing vim.Datacenter instance, None otherwise

        """
        datacenter = pvc.widget.common.choose_datacenter(
            agent=self.agent,
            dialog=self.dialog
        )

        if not datacenter:
            return

        if not datacenter.hostFolder.childEntity:
            self.dialog.msgbox(
                title='Clone Virtual Machine',
                text='No compute resources found in datacenter {}'.format(datacenter.name)
            )
            return

        return datacenter

    def select_cluster(self, folder):
        """
        Select a cluster for Virtual Machine clone placement

        Args:
            folder (vim.Folder): A vim.Folder instance

        Returns:
            A vim.ClusterComputeResource managed entity upon successufuly
            selecting an existing cluster, None otherwise

        """
        cluster = pvc.widget.common.choose_cluster(
            agent=self.agent,
            dialog=self.dialog,
            folder=folder
        )

        if not cluster:
            return

        if not cluster.host:
            self.dialog.msgbox(
                title='Clone Virtual Machine',
                text='No valid hosts found in cluster {}'.format(cluster.name)
            )
            return

        return cluster

    def select_datastore(self, obj):
        """
        Select datastore for Virtual Machine clone placement

        Args:
            obj (vim.ManagedEntity): A Managed Entity containing datastores

        Returns:
            A vim.Datastore managed entity upon successfully
            selecting a datastore, None otherwise

        """
        if not obj.datastore:
            self.dialog.msgbox(
                title='Clone Virtual Machine',
                text='No datastores found on {}'.format(obj.name)
            )
            return

        datastore = pvc.widget.common.choose_datastore(
            agent=self.agent,
            dialog=self.dialog,
            obj=obj
        )

        return datastore



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Cluster Widgets

"""

import pyVmomi
import humanize

import pvc.widget.alarm
import pvc.widget.common
import pvc.widget.checklist
import pvc.widget.debug
import pvc.widget.event
import pvc.widget.form
import pvc.widget.gauge
import pvc.widget.menu
import pvc.widget.performance
import pvc.widget.virtualmachine

__all__ = [
    'ClusterWidget', 'ClusterActionWidget',
    'ClusterHostWidget', 'ClusterVirtualMachineWidget',
]


class ClusterWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Cluster Widget

        Args:
            agent                  (VConnector): A VConnector instance
            dialog              (dialog.Dialog): A Dialog instance
            obj    (vim.ClusterComputeResource): A ClusterComputeResource managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Summary',
                description='General information',
                on_select=self.summary
            ),
            pvc.widget.menu.MenuItem(
                tag='Resources',
                description='Resource usage information',
                on_select=self.resources
            ),
            pvc.widget.menu.MenuItem(
                tag='Actions',
                description='Available Actions',
                on_select=ClusterActionWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Hosts',
                description='Manage hosts in cluster',
                on_select=ClusterHostWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Virtual Machines',
                description='Manage Virtual Machines in cluster',
                on_select=ClusterVirtualMachineWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Datastores',
                description='Manage datastores',
                on_select=pvc.widget.common.datastore_menu,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Networks',
                description='Networking',
                on_select=pvc.widget.common.network_menu,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Performance',
                description='Performance Metrics',
                on_select=pvc.widget.performance.PerformanceProviderWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Events',
                description='View Events',
                on_select=pvc.widget.event.EventWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Alarms',
                description='View triggered alarms',
                on_select=pvc.widget.common.alarm_menu,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='Debug',
                description='Start a Python REPL console',
                on_select=pvc.widget.debug.DebugWidget,
                on_select_args=(locals(), globals())
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select item from menu'
        )

        menu.display()

    def summary(self):
        """
        Cluster general information

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        elements = [
            pvc.widget.form.FormElement(
                label='Hosts',
                item=str(self.obj.summary.numHosts)
            ),
            pvc.widget.form.FormElement(
                label='DRS Enabled',
                item=str(self.obj.configuration.drsConfig.enabled)
            ),
            pvc.widget.form.FormElement(
                label='DRS Mode',
                item=self.obj.configuration.drsConfig.defaultVmBehavior
            ),
            pvc.widget.form.FormElement(
                label='vMotion Migrations',
                item=str(self.obj.summary.numVmotions)
            ),
            pvc.widget.form.FormElement(
                label='Total CPU Cores',
                item=str(self.obj.summary.numCpuCores)
            ),
            pvc.widget.form.FormElement(
                label='Total CPU Threads',
                item=str(self.obj.summary.numCpuThreads)
            ),
            pvc.widget.form.FormElement(
                label='Total CPU Resources',
                item='{} MHz'.format(self.obj.summary.totalCpu)
            ),
            pvc.widget.form.FormElement(
                label='Total Memory',
                item=humanize.naturalsize(self.obj.summary.totalMemory, binary=True)
            ),
            pvc.widget.form.FormElement(
                label='Overall Status',
                item=self.obj.overallStatus
            ),
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title=self.title,
            text='Cluster general information'
        )

        form.display()

    def resources(self):
        """
        Resource usage information

        """
        text = (
            'Not implemented yet.\n'
            'See https://github.com/vmware/pyvmomi/issues/229 '
            'for more details.\n'
        )

        self.dialog.msgbox(
            title=self.title,
            text=text
        )


class ClusterActionWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Cluster Actions Widget

        Args:
            agent                  (VConnector): A VConnector instance
            dialog              (dialog.Dialog): A Dialog instance
            obj    (vim.ClusterComputeResource): A ClusterComputeResource managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Rename',
                description='Rename cluster',
                on_select=pvc.widget.common.rename,
                on_select_args=(self.obj, self.dialog)
            ),
            pvc.widget.menu.MenuItem(
                tag='Remove',
                description='Remove cluster',
                on_select=pvc.widget.common.remove,
                on_select_args=(self.obj, self.agent)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()


class ClusterHostWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Cluster Host Widget

        Args:
            agent                  (VConnector): A VConnector instance
            dialog              (dialog.Dialog): A Dialog instance
            obj    (vim.ClusterComputeResource): A ClusterComputeResource managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Connect',
                description='Connect host to cluster',
                on_select=self.connect_host
            ),
            pvc.widget.menu.MenuItem(
                tag='Disconnect',
                description='Disconnect host(s) from cluster',
                on_select=self.disconnect_host
            ),
            pvc.widget.menu.MenuItem(
                tag='Reconnect',
                description='Reconnect host(s) to cluster',
                on_select=self.reconnect_host
            ),
            pvc.widget.menu.MenuItem(
                tag='View',
                description='View hosts in cluster',
                on_select=pvc.widget.common.host_menu,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select action to be performed'
        )

        menu.display()

    def connect_host(self):
        """
        Connect a host to the cluster and add it to the inventory

        """
        text = (
            'Enter hostname or IP address of the '
            'host to be connected to cluster {}\n'
        )

        elements = [
            pvc.widget.form.FormElement(label='Hostname'),
            pvc.widget.form.FormElement(label='SSL Thumbprint'),
            pvc.widget.form.FormElement(label='Username'),
            pvc.widget.form.FormElement(label='Password', attributes=0x1),
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            mixed_form=True,
            title=self.title,
            text=text.format(self.obj.name)
        )

        code, fields = form.display()

        if code in (self.dialog.CANCEL, self.dialog.ESC):
            return

        if not all(fields.values()):
            self.dialog.msgbox(
                title=self.title,
                text='Invalid input provided'
            )
            return

        connect_spec = pyVmomi.vim.HostConnectSpec(
            hostName=fields['Hostname'],
            sslThumbprint=fields['SSL Thumbprint'],
            userName=fields['Username'],
            password=fields['Password']
        )

        task = self.obj.AddHost(
            spec=connect_spec,
            asConnected=True
        )

        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Connecting {} to cluster ...'.format(fields['Hostname'])
        )

        gauge.display()

    def disconnect_host(self):
        """
        Disconnect host(s) from the cluster

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        items = [
            pvc.widget.checklist.CheckListItem(tag=h.name, description=h.runtime.connectionState)
            for h in self.obj.host if h.runtime.connectionState == pyVmomi.vim.HostSystemConnectionState.connected
        ]

        if not items:
            self.dialog.msgbox(
                title=self.title,
                text='There are no hosts connected to cluster'
            )
            return

        checklist = pvc.widget.checklist.CheckList(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select host(s) to be disconnected from the cluster'
        )

        checklist.display()
        selected = checklist.selected()

        if not selected:
            return

        text = (
            'The following host(s) will be disconnected from cluster.'
            '\n\n{}\n\n'
            'Disconnect host(s) from cluster?\n'
        )
        code = self.dialog.yesno(
            title=self.title,
            text=text.format('\n'.join(selected))
        )

        if code in (self.dialog.ESC, self.dialog.CANCEL):
            return

        host_objects = [h for h in self.obj.host if h.name in selected]
        for host_obj in host_objects:
            task = host_obj.Disconnect()
            gauge = pvc.widget.gauge.TaskGauge(
                title=self.title,
                text='Disconnecting {} from cluster ...'.format(host_obj.name),
                dialog=self.dialog,
                task=task
            )
            gauge.display()

    def reconnect_host(self):
        """
        Reconnect disconnected hosts to cluster

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        items = [
            pvc.widget.checklist.CheckListItem(tag=h.name, description=h.runtime.connectionState)
            for h in self.obj.host if h.runtime.connectionState == pyVmomi.vim.HostSystemConnectionState.disconnected
        ]

        if not items:
            self.dialog.msgbox(
                title=self.title,
                text='There are no disconnected hosts in the cluster'
            )
            return

        checklist = pvc.widget.checklist.CheckList(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select host(s) to be reconnected to the cluster'
        )

        checklist.display()
        selected_hosts = checklist.selected()

        if not selected_hosts:
            return

        host_objects = [h for sh in selected_hosts for h in self.obj.host if sh == h.name]
        for host_obj in host_objects:
            task = host_obj.Reconnect()
            gauge = pvc.widget.gauge.TaskGauge(
                title=self.title,
                text='Reconnecting {} to cluster ...'.format(host_obj.name),
                dialog=self.dialog,
                task=task
            )
            gauge.display()


class ClusterVirtualMachineWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Widget for managing virtual machines in a cluster

        Args:
            agent                  (VConnector): A VConnector instance
            dialog              (dialog.Dialog): A Dialog instance
            obj    (vim.ClusterComputeResource): A vim.ClusterComputeResource instance

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Create',
                description='Create new Virtual Machine',
                on_select=pvc.widget.virtualmachine.CreateVirtualMachineWidget,
                on_select_args=(self.agent, self.dialog, self.obj.parent.parent, self.obj)
            ),
            pvc.widget.menu.MenuItem(
                tag='View',
                description='Virtual Machines in cluster',
                on_select=self.virtual_machine_menu
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()

    def virtual_machine_menu(self):
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        view = self.agent.get_container_view(
            obj_type=[pyVmomi.vim.VirtualMachine],
            container=self.obj
        )
        properties = self.agent.collect_properties(
            view_ref=view,
            obj_type=pyVmomi.vim.VirtualMachine,
            path_set=['name', 'runtime.powerState'],
            include_mors=True
        )
        view.DestroyView()

        if not properties:
            self.dialog.msgbox(
                title=self.title,
                text='No virtual machines found in cluster'
            )
            return

        items = [
            pvc.widget.menu.MenuItem(
                tag=vm['name'],
                description=vm['runtime.powerState'],
                on_select=pvc.widget.virtualmachine.VirtualMachineWidget,
                on_select_args=(self.agent, self.dialog, vm['obj'])
            ) for vm in properties
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select a virtual machine from the menu'
        )

        menu.display()



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Alarm Widgets

"""

import pvc.widget.debug
import pvc.widget.form
import pvc.widget.menu

__all__ = ['AlarmWidget']


class AlarmWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Alarm Widget

        Args:
            agent      (VConnector): A VConnector instance
            dialog  (dialog.Dialog): A Dialog instance
            obj    (vim.AlarmState): A vim.AlarmState instance

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.entity.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Details',
                description='View Alarm Details',
                on_select=self.details
            ),
            pvc.widget.menu.MenuItem(
                tag='Acknowledge',
                description='Acknowledge Alarm',
                on_select=self.acknowledge
            ),
            pvc.widget.menu.MenuItem(
                tag='Reset',
                description='Reset Alarm'
            ),
            pvc.widget.menu.MenuItem(
                tag='Debug',
                description='Start a Python REPL console',
                on_select=pvc.widget.debug.DebugWidget,
                on_select_args=(locals(), globals())
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()

    def details(self):
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        elements = [
            pvc.widget.form.FormElement(
                label='Entity',
                item=self.obj.entity.name
            ),
            pvc.widget.form.FormElement(
                label='Status',
                item=self.obj.overallStatus
            ),
            pvc.widget.form.FormElement(
                label='Name',
                item=self.obj.alarm.info.name
            ),
            pvc.widget.form.FormElement(
                label='Triggered',
                item=str(self.obj.time)
            ),
            pvc.widget.form.FormElement(
                label='Acknowledged',
                item=str(self.obj.acknowledged)
            ),
            pvc.widget.form.FormElement(
                label='Acknowledged At',
                item=str(self.obj.acknowledgedTime) if self.obj.acknowledgedTime else ''
            ),
            pvc.widget.form.FormElement(
                label='Acknowledged By',
                item=self.obj.acknowledgedByUser if self.obj.acknowledgedByUser else ''
            )
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title=self.title,
            text='Alarm details'
        )

        return form.display()

    def acknowledge(self, alarm):
        """
        Acknowledge alarm

        Args:
            alarm (vim.AlarmState): A vim.AlarmState instance

        """
        self.dialog.infobox(
            title=self.title,
            text='Acknowledging alarm ...'
        )

        am = self.agent.si.content.alarmManager
        am.AcknowledgeAlarm(
            alarm=alarm.alarm,
            entity=alarm.entity
        )



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Performance Metrics Widgets

"""

import os
import datetime
import tempfile
import subprocess

import pyVmomi

import pvc.widget.menu
import pvc.widget.form
import pvc.widget.checklist
import pvc.widget.radiolist

__all__ = [
    'PerformanceProviderWidget', 'PerformanceGroupWidget',
    'PerformanceCounterInGroupWidget', 'PerformanceCounterWidget',
    'PerformanceCounterGraphWidget',
]


class PerformanceProviderWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Performance Provider Widget

        Args:
            agent         (VConnector): A VConnector instance
            dialog     (dialog.Dialog): A Dialog instance
            obj    (vim.ManagedEntity): A managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.pm = self.agent.si.content.perfManager
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Summary',
                description='Performance provider summary',
                on_select=self.summary
            ),
            pvc.widget.menu.MenuItem(
                tag='Groups',
                description='Performance counter groups',
                on_select=PerformanceGroupWidget,
                on_select_args=(self.agent, self.dialog, self.obj)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an item from the menu',
        )

        menu.display()

    def summary(self):
        """
        Performance provider summary information

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        provider_summary = self.pm.QueryPerfProviderSummary(
            entity=self.obj
        )

        elements = [
            pvc.widget.form.FormElement(
                label='Real-time statistics support',
                item=str(provider_summary.currentSupported)
            ),
            pvc.widget.form.FormElement(
                label='Historical statistics support',
                item=str(provider_summary.summarySupported)
            ),
            pvc.widget.form.FormElement(
                label='Refresh rate (seconds)',
                item=str(provider_summary.refreshRate)
            )
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title=self.title,
            text='Performance provider summary information'
        )

        form.display()


class PerformanceGroupWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Displays a menu of available counter groups for the provider

        Args:
            agent         (VConnector): A VConnector instance
            dialog     (dialog.Dialog): A Dialog instance
            obj    (vim.ManagedEntity): A managed entity

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.pm = self.agent.si.content.perfManager
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Real-time',
                description='Real-time performance metrics',
                on_select=self.realtime_counter_groups
            ),
            pvc.widget.menu.MenuItem(
                tag='Historical',
                description='Historical performance metrics',
                on_select=self.historical_counter_groups
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an item from the menu',
        )

        menu.display()

    def realtime_counter_groups(self):
        """
        Available real-time counter groups

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        provider_summary = self.pm.QueryPerfProviderSummary(
            entity=self.obj
        )

        if not provider_summary.currentSupported:
            self.dialog.msgbox(
                title=self.title,
                text='Provider does not support real-time statistics'
            )
            return

        metric_id = self.pm.QueryAvailablePerfMetric(
            entity=self.obj,
            intervalId=provider_summary.refreshRate
        )

        if not metric_id:
            self.dialog.msgbox(
                title=self.title,
                text='Performance data is currently not available for entity'
            )
            return

        perf_counter = self.pm.perfCounter
        counters = [c for c in perf_counter for m in metric_id if c.key == m.counterId]
        groups = set([(c.groupInfo.key, c.groupInfo.label) for c in counters])

        items = [
            pvc.widget.menu.MenuItem(
                tag=key,
                description=label,
                on_select=PerformanceCounterInGroupWidget,
                on_select_args=(self.agent, self.dialog, self.obj, metric_id, label, True)
            ) for key, label in groups
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select a performance counter group'
        )

        menu.display()

    def historical_counter_groups(self):
        """
        Available historical counter groups

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        provider_summary = self.pm.QueryPerfProviderSummary(
            entity=self.obj
        )

        if not provider_summary.summarySupported:
            self.dialog.msgbox(
                title=self.title,
                text='Provider does not support historical statistics'
            )
            return

        metric_id = self.pm.QueryAvailablePerfMetric(
            entity=self.obj
        )

        if not metric_id:
            self.dialog.msgbox(
                title=self.title,
                text='Performance data is currently not available for entity'
            )
            return

        perf_counter = self.pm.perfCounter
        counters = [c for c in perf_counter for m in metric_id if c.key == m.counterId]
        groups = set([(c.groupInfo.key, c.groupInfo.label) for c in counters])

        items = [
            pvc.widget.menu.MenuItem(
                tag=key,
                description=label,
                on_select=PerformanceCounterInGroupWidget,
                on_select_args=(self.agent, self.dialog, self.obj, metric_id, label, False)
            ) for key, label in groups
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select a performance counter group'
        )

        menu.display()


class PerformanceCounterInGroupWidget(object):
    def __init__(self, agent, dialog, obj, metric_id, label, realtime):
        """
        Display a menu of available performance counters
        in specific performance group

        Args:
            agent         (VConnector): A VConnector instance
            dialog     (dialog.Dialog): A Dialog instance
            obj    (vim.ManagedEntity): A managed entity
            metric_id           (list): A list of vim.PerformanceManager.MetricId instances,
                                        retrieved by a previous
                                        vim.PerformanceManager.QueryAvailablePerfMetric() call
            label                (str): Performance counter group label
            realtime            (bool): A flag indicating that this counter group contains
                                        real-time counters if set to True. Otherwise consider
                                        counters from this group as historical.

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.pm = self.agent.si.content.perfManager
        self.metric_id = metric_id
        self.label = label
        self.realtime = realtime
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        # Get the unique metrics so we don't get duplicate
        # entries in the resulting menu. Duplicate entries
        # may be seen by counters with more than one
        # object instance, e.g. vmnic0, vmnic1, etc.
        unique_metrics = set([m.counterId for m in self.metric_id])

        perf_counter = self.pm.perfCounter
        counters = [c for c in perf_counter for m in unique_metrics if c.key == m and c.groupInfo.label == self.label]

        items = [
            pvc.widget.menu.MenuItem(
                tag='{0}.{1}.{2}'.format(c.groupInfo.key, c.nameInfo.key, c.unitInfo.key),
                description=c.nameInfo.label,
                on_select=PerformanceCounterWidget,
                on_select_args=(self.agent, self.dialog, self.obj, c, self.realtime)
            ) for c in counters
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text="Performance counters in group '{}'".format(self.label)
        )

        menu.display()


class PerformanceCounterWidget(object):
    def __init__(self, agent, dialog, obj, counter, realtime):
        """
        Performance Counter Widget

        Args:
            agent                           (VConnector): A VConnector instance
            dialog                       (dialog.Dialog): A Dialog instance
            obj                      (vim.ManagedEntity): A managed entity
            counter (vim.PerformanceManager.CounterInfo): A CounterInfo instance
            realtime                              (bool): A flag indicating that this
                                                          counter has been retrieved
                                                          from a real-time counter group if
                                                          set to True. Otherwise we consider
                                                          this counter as retrieved from a
                                                          historical counter group

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.counter = counter
        self.realtime = realtime
        self.pm = self.agent.si.content.perfManager
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Info',
                description='Counter information',
                on_select=self.info
            ),
            pvc.widget.menu.MenuItem(
                tag='Graph',
                description='Display graph',
                on_select=PerformanceCounterGraphWidget,
                on_select_args=(self.agent, self.dialog, self.obj, self.counter, self.realtime)
            ),
        ]

        title = 'Performance counter {0}.{1}.{2}'.format(
            self.counter.groupInfo.key,
            self.counter.nameInfo.key,
            self.counter.unitInfo.key
        )

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text=title
        )

        menu.display()

    def info(self):
        """
        Display information about a counter

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        counter_name = '{0}.{1}.{2}'.format(
            self.counter.groupInfo.key,
            self.counter.nameInfo.key,
            self.counter.unitInfo.key
        )
        intervals = [i.name for i in self.pm.historicalInterval if self.counter.level == i.level]

        elements = [
            pvc.widget.form.FormElement(
                label='Key',
                item=str(self.counter.key)
            ),
            pvc.widget.form.FormElement(
                label='Counter',
                item=counter_name
            ),
            pvc.widget.form.FormElement(
                label='Label',
                item=self.counter.nameInfo.label
            ),
            pvc.widget.form.FormElement(
                label='Description',
                item=self.counter.nameInfo.summary
            ),
            pvc.widget.form.FormElement(
                label='Group',
                item=self.counter.groupInfo.label
            ),
            pvc.widget.form.FormElement(
                label='Unit',
                item=self.counter.unitInfo.label
            ),
            pvc.widget.form.FormElement(
                label='Intervals',
                item=', '.join(intervals)
            ),
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title=self.title,
            text='Performance counter information'
        )

        form.display()


class PerformanceCounterGraphWidget(object):
    def __init__(self, agent, dialog, obj, counter, realtime):
        """
        Widget to plot a gnuplot(1) graph of a performance counter

        Args:
            agent                           (VConnector): A VConnector instance
            dialog                       (dialog.Dialog): A Dialog instance
            obj                      (vim.ManagedEntity): A managed entity
            counter (vim.PerformanceManager.CounterInfo): A CounterInfo instance
            realtime                              (bool): A flag indicating that this
                                                          counter has been retrieved
                                                          from a real-time counter group if
                                                          set to True. Otherwise we consider
                                                          this counter as retrieved from a
                                                          historical counter group

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.counter = counter
        self.realtime = realtime
        self.pm = self.agent.si.content.perfManager
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def display(self):
        try:
            subprocess.Popen(
                args=['gnuplot', '--version'],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE
            )
        except OSError as e:
            self.dialog.msgbox(
                title=self.title,
                text='Unable to find gnuplot(1): \n{}\n'.format(e)
            )
            return

        selected_instances = self.select_counter_instances()
        if not selected_instances:
            self.dialog.msgbox(
                title=self.title,
                text='No counter instances selected'
            )
            return

        metric_id = [
            pyVmomi.vim.PerformanceManager.MetricId(
                counterId=self.counter.key,
                instance='' if instance == self.obj.name else instance
            ) for instance in selected_instances
        ]

        fd, datafile = tempfile.mkstemp(prefix='pvcgnuplot-data-')
        script = self.create_gnuplot_script(
            datafile=datafile,
            instances=selected_instances
        )

        if self.realtime:
            self.realtime_graph(metric_id, datafile, script)
        else:
            self.historical_graph(metric_id, datafile, script)

        os.unlink(datafile)
        os.unlink(script)

    def save_performance_samples(self, path, data):
        """
        Save performance samples to a file

        New samples are appended to the file

        NOTE: If the performance counter unit is percentage we need
              to make sure that the sample value is divided by
              a hundred, as the returned sample value
              represents a 1/100th of the percent.

        Args:
            path                                      (str): Path to the datafile
            data  (vim.PerformanceManager.EntityMetricBase): The data to be saved

        """
        all_values = [v.value for v in data.value]
        samples = zip(data.sampleInfo, *all_values)

        with open(path, 'a') as f:
            for sample in samples:
                timestamp, values = sample[0].timestamp, sample[1:]
                if self.counter.unitInfo.key == 'percent':
                    f.write('{},{}\n'.format(str(timestamp), ','.join([str(v / 100) for v in values])))
                else:
                    f.write('{},{}\n'.format(str(timestamp), ','.join([str(v) for v in values])))

    def create_gnuplot_script(self, datafile, instances):
        """
        Create a gnuplot(1) script for plotting a graph

        Args:
            datafile   (str): Path to a datafile containing performance samples
            instances (list): A list of object instances present in the
                              retrieved performance samples, e.g. vmnic0, vmnic1, etc.

        Returns:
            Path to the created gnuplot(1) script

        """
        gnuplot_term = os.environ['GNUPLOT_TERM'] if os.environ.get('GNUPLOT_TERM') else 'dumb'

        lines = []
        for index, instance in enumerate(instances):
            l = '"{datafile}" using 1:{index} title "{instance}" with lines'.format(
                datafile=datafile,
                index=index+2,
                instance=instance
            )
            lines.append(l)

        script_template = (
            "# gnuplot(1) script created by PVC\n"
            "set title '{name} - {title}'\n"
            "set term {term}\n"
            "set grid\n"
            "set xdata time\n"
            "set timefmt '%Y-%m-%d %H:%M:%S+00:00'\n"
            "# set format x '%H:%M:%S'\n"
            "set xlabel 'Time'\n"
            "set ylabel '{unit}'\n"
            "set key outside right center\n"
            "set datafile separator ','\n"
            "set autoscale fix\n"
            "set yrange [{yrange}]\n"
            "plot {lines}\n"
            "pause {pause}\n"
        )

        # Real-time counters
        # Append any additional gnuplot(1) commands here
        # for real-time counters
        if self.realtime:
            provider_summary = self.pm.QueryPerfProviderSummary(
                entity=self.obj
            )
            pause = provider_summary.refreshRate

            # Append the 'reread' gnuplot(1) command
            script_template += 'reread\n'
        else:
            # Historical counters
            pause = -1

        # Set a yrange for counters which unit is percentage
        yrange = '0:100' if self.counter.unitInfo.key == 'percent' else ''

        gnuplot_script = script_template.format(
            name=self.obj.name,
            title=self.counter.nameInfo.summary,
            term=gnuplot_term,
            unit=self.counter.unitInfo.label,
            lines=', '.join(lines),
            pause=pause,
            yrange=yrange
        )

        fd, path = tempfile.mkstemp(prefix='pvcgnuplot-script-')
        with open(path, 'w') as f:
            f.write(gnuplot_script)

        return path

    def select_counter_instances(self):
        """
        Prompts the user to select counter instances

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        if self.realtime:
            provider_summary = self.pm.QueryPerfProviderSummary(
                entity=self.obj
            )
            interval_id = provider_summary.refreshRate
        else:
            interval_id = None

        metric_id = self.pm.QueryAvailablePerfMetric(
            entity=self.obj,
            intervalId=interval_id
        )
        metrics = [m for m in metric_id if m.counterId == self.counter.key]
        instances = [m.instance if m.instance else self.obj.name for m in metrics]
        items = [
            pvc.widget.checklist.CheckListItem(tag=instance)
            for instance in instances
        ]
        checklist_text = 'Select instances for counter {0}.{1}.{2}'.format(
            self.counter.groupInfo.key,
            self.counter.nameInfo.key,
            self.counter.unitInfo.key
        )

        checklist = pvc.widget.checklist.CheckList(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text=checklist_text,
        )
        checklist.display()

        return checklist.selected()

    def select_historical_interval(self):
        """
        Prompts the user to select an existing historical interval

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        intervals = [i.name for i in self.pm.historicalInterval]
        items = [
            pvc.widget.radiolist.RadioListItem(tag=interval) for interval in intervals
        ]

        radiolist = pvc.widget.radiolist.RadioList(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select a historical performance interval',
        )

        return radiolist.display()

    def realtime_graph(self, metric_id, datafile, script):
        """
        Plot a real-time graph

        Args:
            metric_id (list): A list of vim.PerformanceManager.MetricId instances
            datafile   (str): Path to a datafile to store collected samples
            script     (str): Path to a gnuplot(1) script used to plot the graph

        """
        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        provider_summary = self.pm.QueryPerfProviderSummary(
            entity=self.obj
        )
        interval_id = provider_summary.refreshRate

        # Query spec to get data from the past hour
        one_hour_ago = self.agent.si.CurrentTime() - datetime.timedelta(seconds=3600)
        query_spec_last_hour = pyVmomi.vim.PerformanceManager.QuerySpec(
            entity=self.obj,
            metricId=metric_id,
            intervalId=interval_id,
            startTime=one_hour_ago
        )
        data = self.pm.QueryPerf(querySpec=[query_spec_last_hour]).pop()
        self.save_performance_samples(
            path=datafile,
            data=data
        )

        # Query spec used to continuously get new performance data
        query_spec = pyVmomi.vim.PerformanceManager.QuerySpec(
            maxSample=1,
            entity=self.obj,
            metricId=metric_id,
            intervalId=interval_id
        )

        p = subprocess.Popen(
            args=['gnuplot', script]
        )

        text = (
            'Graph updates every {} seconds.\n\n'
            'Press CANCEL in order to stop plotting the '
            'graph and exit.\n'
        )

        while True:
            data = self.pm.QueryPerf(querySpec=[query_spec]).pop()
            self.save_performance_samples(
                path=datafile,
                data=data
            )
            code = self.dialog.pause(
                title=self.title,
                text=text.format(interval_id),
                height=15,
                width=60,
                seconds=interval_id
            )
            if code == self.dialog.CANCEL:
                break

        p.terminate()

    def historical_graph(self, metric_id, datafile, script):
        """
        Plot a historical graph

        Args:
            metric_id (list): A list of vim.PerformanceManager.MetricId instances
            datafile   (str): Path to a datafile to store collected samples
            script     (str): Path to a gnuplot(1) script used to plot the graph

        """
        code, interval = self.select_historical_interval()
        if code in (self.dialog.CANCEL, self.dialog.ESC) or not interval:
            return

        self.dialog.infobox(
            title=self.title,
            text='Retrieving information ...'
        )

        interval_id = [i.samplingPeriod for i in self.pm.historicalInterval if i.name == interval].pop()
        query_spec = pyVmomi.vim.PerformanceManager.QuerySpec(
            entity=self.obj,
            metricId=metric_id,
            intervalId=interval_id
        )
        data = self.pm.QueryPerf(querySpec=[query_spec]).pop()
        self.save_performance_samples(
            path=datafile,
            data=data
        )

        p = subprocess.Popen(
            args=['gnuplot', script]
        )

        p.wait()



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Form widgets

"""

__all__ = ['Form', 'FormElement']


class FormElement(object):
    def __init__(self, label, item='', field_length=60, input_length=0, attributes=0x0):
        """
        A form element

        Args:
           label       (str): Label of the element
           item        (str): Initial value for the element
           field_lenght(int): Number of characters used to display the item
           field_input (int): Input field characters
           attributes  (int): A bitmask used for elements used in mixed forms

        """
        self.label = label
        self.item = item
        self.field_length = field_length
        self.input_length = input_length
        self.attributes = attributes


class Form(object):
    def __init__(self, dialog, form_elements, mixed_form=False, **kwargs):
        """
        A form widget

        Args:
            dialog        (dialog.Dialog): A Dialog instance
            form_elements          (list): A list of FormElement instances to display
            mixed_form             (bool): If True then create a mixed form, otherwise
                                           create a standard form
            kwargs                 (dict): Additional args to be passed to dialog(1)

        """
        self.form = None
        self.dialog = dialog
        self.form_elements = form_elements
        self.mixed_form = mixed_form
        self.kwargs = kwargs
        self._labels = [e.label for e in self.form_elements]

        # Starting column at which to display the items' values
        self._xi = len(max(self._labels, key=len)) + 3

        if self.mixed_form:
            self.form = self.dialog.mixedform
            self._elements = [(e.label, row + 1, 1, e.item, row + 1, self._xi, e.field_length, e.input_length, e.attributes) for row, e in enumerate(self.form_elements)]
        else:
            self.form = self.dialog.form
            self._elements = [(e.label, row + 1, 1, e.item, row + 1, self._xi, e.field_length, e.input_length) for row, e in enumerate(self.form_elements)]

    def display(self):
        code, items = self.form(
            elements=self._elements,
            **self.kwargs
        )

        result = (code, {l: i for l, i in zip(self._labels, items)})

        return result



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Message Of The Day module

"""

import os
import tempfile

__all__ = ['MOTDWidget']


class MOTDWidget(object):
    def __init__(self, agent, dialog):
        """
        Message Of The Day Widget

        Args:
            agent     (VConnector): A VConnector instance
            dialog (dialog.Dialog): A Dialog instance

        """
        self.agent = agent
        self.dialog = dialog
        self.display()

    def display(self):
        self.dialog.infobox(
            text='Retrieving information ...'
        )

        sm = self.agent.si.content.sessionManager
        motd = sm.message + '\n'
        fd, motd_file = tempfile.mkstemp(prefix='pvc_motd_')

        with open(motd_file, 'w') as f:
            f.write(motd)

        code, text = self.dialog.editbox(
            title='Message Of The Day',
            filepath=motd_file
        )
        os.unlink(motd_file)

        if code == self.dialog.OK:
            self.dialog.infobox(
                text='Setting message of the day ...'
            )
            sm.UpdateServiceMessage(message=text.rstrip())



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
VNC module

"""

import os
import random
import socket
import string
import time
import tempfile

import pyVmomi
import pvc.widget.form
import pvc.widget.menu
import pvc.widget.gauge

from subprocess import Popen, PIPE

__all__ = ['VncWidget']


class VncWidget(object):
    def __init__(self, dialog, obj):
        """
        VNC widget

        Args:
            dialog      (dialog.Dialog): A Dialog instance
            obj    (vim.VirtualMachine): A vim.VirtualMachine managed entity

        """
        self.dialog = dialog
        self.obj = obj
        self.title = '{} ({})'.format(self.obj.name, self.obj.__class__.__name__)
        self.display()

    def _port_is_open(self, host, port, timeout=3.0):
        """
        Probes a port to check if it is open or not

        Returns:
            True if the port is open, False otherwise

        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(timeout)
        err_code = sock.connect_ex((host, port))
        sock.close()

        return not err_code

    def _get_available_port(self, attempts):
        """
        Look for available port to use for the VNC console

        We search for available port on the HostSystem where
        our Virtual Machine is running in the port range 5901-5999.

        Args:
            attempts (int): Perform that many iterations when searching for a port

        Returns:
            A random port in the range 5901-5999 if an available
            port was found, None otherwise

        """
        self.dialog.infobox(
            title=self.title,
            text='Searching for available port ...'
        )

        host = self.obj.runtime.host
        host_ip = host.config.network.vnic[0].spec.ip.ipAddress

        for i in range(attempts):
            random_port = 5900 + random.randrange(1, 99)
            if not self._port_is_open(host=host_ip, port=random_port):
                return random_port

        return None

    def _get_random_password(self, length=8):
        """
        Get a random password

        Args:
            length (int): Length of the returned password

        """
        chars = string.ascii_letters + string.digits

        return ''.join(random.choice(chars) for _ in range(length))

    def _get_extra_config_options(self):
        """
        Get Virtual Machine extra configuration options

        Returns:
            A dictionary of the extra config options

        """
        return {o.key: o.value for o in self.obj.config.extraConfig}

    def _configure_vnc_options(self, enabled, port, password):
        """
        Configure VNC console options

        Args:
            enabled (bool): If True enable the VNC console
            port     (int): VNC port to be configured
            password (str): VNC password to be configured

        """
        options = [
            pyVmomi.vim.OptionValue(key='RemoteDisplay.vnc.enabled', value=str(enabled)),
            pyVmomi.vim.OptionValue(key='RemoteDisplay.vnc.port', value=str(port)),
            pyVmomi.vim.OptionValue(key='RemoteDisplay.vnc.password', value=str(password)),
        ]
        spec = pyVmomi.vim.VirtualMachineConfigSpec(extraConfig=options)

        task = self.obj.ReconfigVM_Task(spec=spec)
        gauge = pvc.widget.gauge.TaskGauge(
            title=self.title,
            text='Configuring VNC Settings',
            dialog=self.dialog,
            task=task
        )
        gauge.display()

    def display(self):
        """
        Main widget method

        """
        items = [
            pvc.widget.menu.MenuItem(
                tag='Console',
                description='Launch VNC Console',
                on_select=self.launch_console
            ),
            pvc.widget.menu.MenuItem(
                tag='Enable',
                description='Enable VNC Console',
                on_select=self.enable_vnc
            ),
            pvc.widget.menu.MenuItem(
                tag='Disable',
                description='Disable VNC Console',
                on_select=self.disable_vnc
            ),
            pvc.widget.menu.MenuItem(
                tag='Settings',
                description='Manual VNC Configuration',
                on_select=self.settings
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=self.title,
            text='Select an action to be performed'
        )

        menu.display()

    def enable_vnc(self):
        """
        Enable VNC Console for the Virtual Machine

        """
        self.dialog.infobox(
            title=self.title,
            text='Enabling VNC Console ...'
        )

        extra_config = self._get_extra_config_options()
        enabled = extra_config.get('RemoteDisplay.vnc.enabled')
        port = extra_config.get('RemoteDisplay.vnc.port')
        password = extra_config.get('RemoteDisplay.vnc.password')

        if str(enabled).lower() == 'true':
            self.dialog.msgbox(
                title=self.title,
                text='VNC console is already enabled'
            )
            return

        port = port if port else self._get_available_port(attempts=10)
        if not port:
            self.dialog.msgbox(
                title=self.title,
                text='No available ports for VNC connection, try again later.'
            )
            return

        password = password if password else self._get_random_password()
        self._configure_vnc_options(
            enabled=True,
            port=port,
            password=password
        )

    def disable_vnc(self):
        """
        Disable VNC console for the Virtual Machine

        """
        options = [
            pyVmomi.vim.OptionValue(key='RemoteDisplay.vnc.enabled', value='false'),
        ]
        spec = pyVmomi.vim.VirtualMachineConfigSpec(extraConfig=options)

        task = self.obj.ReconfigVM_Task(spec=spec)
        gauge = pvc.widget.gauge.TaskGauge(
            dialog=self.dialog,
            task=task,
            title=self.title,
            text='Disabling VNC Console',
        )

        gauge.display()

    def settings(self):
        """
        Manual configuration of the VNC settings

        """
        extra_config = self._get_extra_config_options()
        enabled = extra_config.get('RemoteDisplay.vnc.enabled', 'false')
        port = extra_config.get('RemoteDisplay.vnc.port', '')
        password = extra_config.get('RemoteDisplay.vnc.password', '')

        elements = [
            pvc.widget.form.FormElement(
                label='Enabled',
                item=enabled
            ),
            pvc.widget.form.FormElement(
                label='Port',
                item=port
            ),
            pvc.widget.form.FormElement(
                label='Password',
                item=password,
                input_length=8
            )
        ]

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title=self.title,
            text='VNC Configuration Options'
        )

        code, fields = form.display()
        if not all(fields.values()):
            self.dialog.msgbox(
                title=self.title,
                text='Invalid configuration settings'
            )
            return

        if code == self.dialog.OK:
            self._configure_vnc_options(
                enabled=fields['Enabled'],
                port=fields['Port'],
                password=fields['Password']
            )

    def launch_console(self):
        """
        Launch a VNC Console

        """
        if self.obj.runtime.powerState != pyVmomi.vim.VirtualMachinePowerState.poweredOn:
            self.dialog.msgbox(
                title=self.title,
                text='You need to power on the Virtual Machine first'
            )
            return

        extra_config = self._get_extra_config_options()
        enabled = extra_config.get('RemoteDisplay.vnc.enabled')
        port = extra_config.get('RemoteDisplay.vnc.port')
        password = extra_config.get('RemoteDisplay.vnc.password')

        if str(enabled).lower() != 'true':
            self.dialog.msgbox(
                title=self.title,
                text='VNC console is disabled, enable the console first'
            )
            return

        host = self.obj.runtime.host
        host_ip = host.config.network.vnic[0].spec.ip.ipAddress

        self.dialog.infobox(
            title=self.title,
            text='Launching console ...'
        )

        if not self._port_is_open(host=host_ip, port=int(port)):
            text = (
                'Host {} with IP address {} is not reachable on port {}\n'
                'Cannot establish a connection to the Virtual Machine console\n'
            )
            self.dialog.msgbox(
                title=self.title,
                text=text.format(host.name, host_ip, port)
            )
            return

        try:
            fd, vncpasswd_file = tempfile.mkstemp(prefix='pvcvnc_')
            with Popen(args=['vncpasswd', '-f'], stdin=PIPE, stdout=PIPE, stderr=PIPE) as p:
                data = '{0}\n{0}\n'.format(password)
                out, err = p.communicate(bytes(data, 'utf-8'))
            with open(vncpasswd_file, 'wb') as f:
                f.write(out)
        except OSError as e:
            self.dialog.msgbox(
                title=self.title,
                text='Cannot create a vncpasswd(1) file: \n{}\n'.format(e)
            )
            return

        try:
            Popen(
                args=['vncviewer', '-passwd', vncpasswd_file, '{}:{}'.format(host_ip, port)],
                stdout=PIPE,
                stderr=PIPE
            )
        except OSError as e:
            self.dialog.msgbox(
                title=self.title,
                text='Cannot launch console: \n{}\n'.format(e)
            )

        time.sleep(3)
        os.unlink(vncpasswd_file)



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Home Widget

"""

import pvc.widget.menu
import pvc.widget.inventory
import pvc.widget.administration

__all__ = ['HomeWidget']


class HomeWidget(object):
    def __init__(self, agent, dialog):
        """
        Home widget

        Args:
            agent (VConnector): A VConnector instance
            dialog    (Dialog): A Dialog instance

        """
        self.agent = agent
        self.dialog = dialog

    def display(self):
        self.warn_if_not_vcenter()
        self.show_motd()

        items = [
            pvc.widget.menu.MenuItem(
                tag='Inventory',
                description='Inventory Menu',
                on_select=pvc.widget.inventory.InventoryWidget,
                on_select_args=(self.agent, self.dialog)
            ),
            pvc.widget.menu.MenuItem(
                tag='Administration',
                description='Administration Menu',
                on_select=pvc.widget.administration.AdministrationWidget,
                on_select_args=(self.agent, self.dialog)
            ),
        ]

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title='Home',
            text='Select an item from the menu',
            cancel_label='Logout'
        )

        menu.display()

    def warn_if_not_vcenter(self):
        about = self.agent.si.content.about

        if about.apiType == 'VirtualCenter':
            return

        text = (
            'You are currently connected to a {} system.\n\n'
            'Some of the features provided by PVC may or may not '
            'be available for the host to which you are currently '
            'connected.\n\n'
            'In order to take full advantage of all PVC '
            'features you should disconnect now and connect to a '
            'VMware vCenter server managing this host.'
        )

        self.dialog.msgbox(
            title='Warning',
            text=text.format(about.fullName)
        )

        view = self.agent.get_host_view()
        host = view.view[0]
        management_ip = host.summary.managementServerIp
        view.DestroyView()

        if management_ip:
            text = (
                'This host is currently being managed by the '
                'VMware vCenter server with IP address {0}.\n\n'
                'You should disconnect now and connect to the '
                'VMware vCenter server at {0}.\n'
            )
            self.dialog.msgbox(
                title='Warning',
                text=text.format(management_ip)
            )

    def show_motd(self):
        sm = self.agent.si.content.sessionManager
        motd = sm.message

        if motd:
            self.dialog.msgbox(
                title='Message Of The Day',
                text=motd
            )



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Session module

"""

import pvc.widget.debug
import pvc.widget.menu
import pvc.widget.form

__all__ = ['SessionWidget']


class SessionWidget(object):
    def __init__(self, agent, dialog, obj):
        """
        Session Widget

        Args:
            agent        (VConnector): A VConnector instance
            dialog    (dialog.Dialog): A Dialog instance
            session (vim.UserSession): A vim.UserSession instance

        """
        self.agent = agent
        self.dialog = dialog
        self.obj = obj
        self.display()

    def display(self):
        items = [
            pvc.widget.menu.MenuItem(
                tag='Details',
                description='View Session Details',
                on_select=self.details
            ),
            pvc.widget.menu.MenuItem(
                tag='Terminate',
                description='Terminate Session',
                on_select=self.terminate
            ),
            pvc.widget.menu.MenuItem(
                tag='Debug',
                description='Start a Python REPL console',
                on_select=pvc.widget.debug.DebugWidget,
                on_select_args=(locals(), globals())
            ),
        ]

        current_session = self.agent.si.content.sessionManager.currentSession
        if current_session.key == self.obj.key:
            title = 'Session {}@{} (This Session)'
        else:
            title = 'Session {}@{}'

        menu = pvc.widget.menu.Menu(
            items=items,
            dialog=self.dialog,
            title=title.format(self.obj.userName, self.obj.ipAddress),
            text='Select an action to be performed'
        )

        menu.display()

    def details(self):
        self.dialog.infobox(
            text='Retrieving information ...'
        )

        elements = [
            pvc.widget.form.FormElement(
                label='Username',
                item=self.obj.userName
            ),
            pvc.widget.form.FormElement(
                label='Full Name',
                item=self.obj.fullName
            ),
            pvc.widget.form.FormElement(
                label='Login Time',
                item=str(self.obj.loginTime)
            ),
            pvc.widget.form.FormElement(
                label='Last Active',
                item=str(self.obj.lastActiveTime)
            ),
            pvc.widget.form.FormElement(
                label='Idle',
                item=str(self.agent.si.CurrentTime() - self.obj.lastActiveTime)
            ),
            pvc.widget.form.FormElement(
                label='IP Address',
                item=self.obj.ipAddress
            ),
            pvc.widget.form.FormElement(
                label='User Agent',
                item=self.obj.userAgent
            ),
            pvc.widget.form.FormElement(
                label='API Invocations',
                item=str(self.obj.callCount)
            ),
        ]

        current_session = self.agent.si.content.sessionManager.currentSession
        if current_session.key == self.obj.key:
            title = 'Session {}@{} (This Session)'
        else:
            title = 'Session {}@{}'

        form = pvc.widget.form.Form(
            dialog=self.dialog,
            form_elements=elements,
            title=title.format(self.obj.userName, self.obj.ipAddress),
            text='Session details'
        )

        form.display()

    def terminate(self):
        current_session = self.agent.si.content.sessionManager.currentSession

        if current_session.key == self.obj.key:
            self.dialog.msgbox(
                text='Cannot terminate current session'
            )
            return

        code, tag = self.dialog.yesno(
            text='Terminate session {}@{}?'.format(self.obj.userName, self.obj.ipAddress)
        )

        if code in (self.dialog.CANCEL, self.dialog.ESC):
            return

        self.dialog.infobox(
            text='Terminating session ...'
        )

        sm = self.agent.si.content.sessionManager
        sm.TerminateSession(
            sessionId=self.obj.key
        )



# Copyright (c) 2015 Marin Atanasov Nikolov <dnaeon@gmail.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer
#    in this position and unchanged.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR(S) ``AS IS'' AND ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
# NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
# THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Menu Widgets

"""

__all__ = ['Menu', 'MenuItem']


class MenuItem(object):
    def __init__(self, tag, description, on_select=None, on_select_args=(), on_select_kwargs={}):
        """
        A menu item

        Args:
            tag                 (str): Unique tag name for the menu item
            description         (str): Short description of the item
            on_select      (callable): A callable to execute when the item is selected
            on_select_args    (tuple): Args to pass to the callable if the item has been selected

        """
        self.tag = tag
        self.description = description
        self.on_select = on_select
        self.on_select_args = on_select_args
        self.on_select_kwargs = on_select_kwargs

        if self.on_select and not callable(self.on_select):
            raise TypeError('Need a callable for item callback')

    def selected(self):
        return self.on_select(*self.on_select_args, **self.on_select_kwargs)


class Menu(object):
    def __init__(self, items, dialog, return_selected=False, **kwargs):
        """
        Menu class

        Args:
            items                    (list): A list of MenuItem instances
            dialog          (dialog.Dialog): A Dialog instance
            return_selected          (bool): If True them just return the selected item
            kwargs                   (dict): Additional args to be passed to dialog(1)

        """
        self.items = items
        self.dialog = dialog
        self.return_selected = return_selected
        self.kwargs = kwargs
        self.choices = [(item.tag, item.description) for item in self.items]
        self._registry = {item.tag: item for item in items}

    def display(self):
        default_item = ''
        while True:
            code, tag = self.dialog.menu(
                choices=self.choices,
                default_item=default_item,
                **self.kwargs
            )

            if code in (self.dialog.CANCEL, self.dialog.ESC):
                return code

            item = self._registry.get(tag)
            default_item = tag

            if self.return_selected:
                return item

            if item.on_select:
                item.selected()
            else:
                self.dialog.msgbox('Not implemented')
